


bool intersectEdgeWithSurface(MyInterval coef, MyInterval qa, MyInterval qb, MyInterval qc,
                              MyInterval qaLeftPower, MyInterval qaRightPower,
                              MyInterval qbLeftPower, MyInterval qbRightPower,
                              MyInterval qcLeftPower, MyInterval qcRightPower,
                              int edge,
                              MyInterval &pointa, MyInterval &pointb, MyInterval &pointc)
// direction = 1 ---  b = coef * \sqrt{a^2 + c^2}
// direction = -1 --- b = -coef * \sqrt{a^2 + c^2}
//       *-------------*
//      /|            /|
//     z |           / |
//    /  *----------/--*
//   *-------------*  /
//   y /           | /
//   |/            |/
//   *-----x-------*
{
    MyInterval temp;
    bool intersect;

    switch (edge)
    {
        // front wall: 0 -- left-front, 1 -- right-front, 2 -- lower-front, 3 -- upper-front
        //       *-------------*
        //      /|            /|
        //     / |           / |
        //    /  *----------/--*
        //   *=====3=======*  /
        //   0 /           1 /
        //   |/            |/
        //   *=====2=======*

    case 0: // intersection with the left-front edge
        pointa = qa.leftBound();
        pointc = qc.leftBound();
        //temp = sqrt(power(MyInterval(pointa), 2) + power(MyInterval(pointc), 2)) * coef;
        temp = sqrt(qaLeftPower + qcLeftPower) * coef;
        intersect = myIntersection(temp, qb, pointb);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qb, pointb);
        }
        //*/
        break;

    case 1: // intersection with the right-front edge
        pointa = qa.rightBound();
        pointc = qc.leftBound();
        //temp = sqrt(power(pointa, 2) + power(pointc, 2)) * coef;
        temp = sqrt(qaRightPower + qcLeftPower) * coef;
        intersect = myIntersection(temp, qb, pointb);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qb, pointb);
        }
        //*/
        break;

    case 2: // intersection with the lower edge
        pointb = qb.leftBound();
        pointc = qc.leftBound();
        //temp = sqrt(power(pointb, 2) + power(pointc, 2)) * coef;
        temp = sqrt(qbLeftPower + qcLeftPower) * coef;
        intersect = myIntersection(temp, qa, pointa);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qa, pointa);
        }
        //*/
        break;

    case 3: // intersection with the upper edge
        pointb = qb.rightBound();
        pointc = qc.leftBound();
        //temp = sqrt(power(pointb, 2) + power(pointc, 2)) * coef;
        temp = sqrt(qbRightPower + qcLeftPower) * coef;
        intersect = myIntersection(temp, qa, pointa);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qa, pointa);
        }
        //*/
        break;

        // back wall: 4 -- left-back, 5 -- right-back, 6 -- lower-back, 7 -- upper-back
        //       *=====7=======*
        //      /4            /5
        //     / |           / |
        //    /  *=====6====/==*
        //   *-------------*  /
        //   | /           | /
        //   |/            |/
        //   *-------------*
    case 4: // intersection with the left-back edge
        pointa = qa.leftBound();
        pointc = qc.rightBound();
        //temp = sqrt(power(pointa, 2) + power(pointc, 2)) * coef;
        temp = sqrt(qaLeftPower + qcRightPower) * coef;
        intersect = myIntersection(temp, qb, pointb);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qb, pointb);
        }
        //*/
        break;

    case 5: // intersection with the right-back edge
        pointa = qa.rightBound();
        pointc = qc.rightBound();
        //temp = sqrt(power(pointa, 2) + power(pointc, 2)) * coef;
        temp = sqrt(qaRightPower + qcRightPower) * coef;
        intersect = myIntersection(temp, qb, pointb);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qb, pointb);
        }
        //*/
        break;

    case 6: // intersection with the lower-back edge
        pointb = qb.leftBound();
        pointc = qc.rightBound();
        //temp = sqrt(power(pointb, 2) + power(pointc, 2)) * coef;
        temp = sqrt(qbLeftPower + qcRightPower) * coef;
        intersect = myIntersection(temp, qa, pointa);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qa, pointa);
        }
        //*/
        break;

    case 7: // intersection with the upper-back edge
        pointb = qb.rightBound();
        pointc = qc.rightBound();
        //temp = sqrt(power(pointb, 2) + power(pointc, 2)) * coef;
        temp = sqrt(qbRightPower + qcRightPower) * coef;
        intersect = myIntersection(temp, qa, pointa);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qa, pointa);
        }
        //*/
        break;

        // left wall:  8 -- lower-left, 9 -- upper-left
        //       *-------------*
        //      /|            /|
        //     9 |           / |
        //    /  *----------/--*
        //   *-------------*  /
        //   | 8           | /
        //   |/            |/
        //   *-------------*
    case 8: // intersection with the left-bottom edge
        pointa = qa.leftBound();
        pointb = qb.leftBound();
        //temp = sqrt(power(pointa, 2) + power(pointb, 2)) * coef;
        temp = sqrt(qaLeftPower + qbLeftPower) * coef;
        intersect = myIntersection(temp, qc, pointc);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qc, pointc);
        }
        //*/
        break;

    case 9: // intersection with the right-top edge
        pointa = qa.leftBound();
        pointb = qb.rightBound();
        //temp = sqrt(power(pointa, 2) + power(pointb, 2)) * coef;
        temp = sqrt(qaLeftPower + qbRightPower) * coef;
        intersect = myIntersection(temp, qc, pointc);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qc, pointc);
        }
        //*/
        break;

        // right wall: 10 -- lower-right, 11 -- upper-right
        //       *-------------*
        //      /|            /|
        //     / |          10 |
        //    /  *----------/--*
        //   *-------------*  /
        //   | /           |11
        //   |/            |/
        //   *-------------*
    case 10: // intersection with the right-bottom edge
        pointa = qa.rightBound();
        pointb = qb.leftBound();
        //temp = sqrt(power(pointa, 2) + power(pointb, 2)) * coef;
        temp = sqrt(qaRightPower + qbLeftPower) * coef;
        intersect = myIntersection(temp, qc, pointc);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qc, pointc);
        }
        //*/
        break;

    case 11: // intersection with the right-top edge
        pointa = qa.rightBound();
        pointb = qb.rightBound();
        //temp = sqrt(power(pointa, 2) + power(pointb, 2)) * coef;
        temp = sqrt(qaRightPower + qbRightPower) * coef;
        intersect = myIntersection(temp, qc, pointc);
        ///*
        if (!intersect)
        {
            temp = -temp;
            intersect = myIntersection(temp, qc, pointc);
        }
        //*/
        break;

    default:
        intersect = false;
        break;
    }

    return intersect;
}


MyInterval radiusXYZn(MyInterval x, MyInterval y, MyInterval z, int exp)
{
    return power(sqrt(power(x, 2) + power(y, 2) + power(z, 2)), exp);
}

/*-------------------*/

void setMinMax(MyInterval x, MyInterval y, MyInterval z,
               MyInterval &qx, MyInterval &qy, MyInterval &qz)
{
    qx = intervalHull(x, qx);
    qy = intervalHull(y, qy);
    qz = intervalHull(z, qz);
}

void tempValues(MyInterval x, MyInterval y, MyInterval z,
                MyInterval &tempx, MyInterval &tempy, MyInterval &tempz,
                int rExp, int bExp)
// tempx = x^bExp/r^rExp
// tempy = y^bExp/r^rExp
// tempz = z^bExp/r^rExp
{
    MyInterval rMinusrExp = radiusXYZn(x, y, z, -rExp);
    tempx = rMinusrExp * power(x, bExp);
    tempy = rMinusrExp * power(y, bExp);
    tempz = rMinusrExp * power(z, bExp);
}



void cuteCalculation(const MyInterval &x, const MyInterval &y, const MyInterval &z,
                     MyInterval &qx, MyInterval &qy, MyInterval &qz,
                     int rExp, int bExp)
/* in the paper: bExp = a, rExp = b
 * we are calculating: q^bExp/|q|^{rExp}
 * qx = x^bExp/r^{rExp}, qy = y^bExp/r^{rExp}, qz = z^bExp/r^{rExp}
 * we consider sufraces:
 * * x = +/- sqrt((y^2 + z^2)a/(b-a))
 * * y = +/- sqrt((x^2 + z^2)a/(b-a))
 * * z = +/- sqrt((x^2 + y^2)a/(b-a))
 * * x = 0
 * * y = 0
 * * z = 0
 */
{
    if (rExp <= bExp)
    {
        cout << "Wrong exponents!";
        exit(2);
    }

    MyInterval r = sqrt(power(x, 2) + power(y, 2)+power(z,2));
    MyInterval r3 = power(r, -rExp);
    if (!testIfDoCute(r)) // we do a naive evaluation
    {
        qx = power(x, bExp) * r3;
        qy = power(y, bExp) * r3;
        qz = power(z, bExp) * r3;
        return;
    }

    MyInterval coef = sqrt(MyInterval(bExp) / (rExp - bExp));
    // coef = a/(b - a)
    MyInterval qMinX, qMaxX, qMinY, qMaxY, qMinZ, qMaxZ, qTemp;
    // first check the corners of the `qi` qube:
    MyInterval xAcc = x.leftBound();
    MyInterval yAcc = y.leftBound();
    MyInterval zAcc = z.leftBound();
    // bottom-front-left corner: (1)
    MyInterval tempx;
    MyInterval tempy;
    MyInterval tempz;
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    qx = tempx;
    qy = tempy;
    qz = tempz;

    // top-front-left corner: (2)
    yAcc = y.rightBound();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // (3)
    zAcc = z.rightBound();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // top-back-right corner: (4)
    xAcc = x.rightBound();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // top-front-right corner: (5)
    zAcc = z.leftBound();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // bottom-front-right corner: (6)
    yAcc = y.leftBound();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // bottom-right corner: (7)
    zAcc = z.rightBound();
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // (8)
    xAcc = x.leftBound(); // 17.09
    tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
    setMinMax(tempx, tempy, tempz, qx, qy, qz);

    // now we have: qx, qy, qz -- min and max from the corners
    MyInterval xI, yI, zI;
    // intersection of the surface `y = coef * \sqrt{x^2 + z^2}` and the edges of the qube `q`:

    MyInterval xLeftPower = power(x.leftBound(), 2);
    MyInterval xRightPower = power(x.rightBound(), 2);
    MyInterval yLeftPower = power(y.leftBound(), 2);
    MyInterval yRightPower = power(y.rightBound(), 2);
    MyInterval zLeftPower = power(z.leftBound(), 2);
    MyInterval zRightPower = power(z.rightBound(), 2);

    for (int edge = 0; edge < 11; edge++)
    {
        if (intersectEdgeWithSurface(coef, x, y, z,
                                     xLeftPower, xRightPower, yLeftPower, yRightPower, zLeftPower, zRightPower,
                                     edge, xI, yI, zI))
        {
            tempValues(xI, yI, zI, tempx, tempy, tempz, rExp, bExp);
            setMinMax(tempx, tempy, tempz, qx, qy, qz);
        }

        if (intersectEdgeWithSurface(coef, y, z, x,
                                     yLeftPower, yRightPower, zLeftPower, zRightPower, xLeftPower, xRightPower,
                                     edge, yI, zI, xI))
        {
            tempValues(xI, yI, zI, tempx, tempy, tempz, rExp, bExp);
            setMinMax(tempx, tempy, tempz, qx, qy, qz);
        }

        if (intersectEdgeWithSurface(coef, z, x, y,
                                     zLeftPower, zRightPower, xLeftPower, xRightPower, yLeftPower, yRightPower,
                                     edge, zI, xI, yI))
        {
            tempValues(xI, yI, zI, tempx, tempy, tempz, rExp, bExp);
            setMinMax(tempx, tempy, tempz, qx, qy, qz);
        }
    }

    // intersection of the surface `y = -coef * \sqrt{x^2 + y^2}` and the edges of the qube `q`:

    // intersection of `x = 0 && z = 0`
    if (isSingular(x) && isSingular(z))
    {
        xAcc = 0.0;
        yAcc = y.leftBound();
        zAcc = 0.0;
        tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);

        yAcc = y.rightBound();
        tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }

    // intersection of `x = 0 && y = 0`
    if (isSingular(x) && isSingular(y))
    {
        xAcc = 0.0;
        yAcc = 0.0;
        zAcc = z.leftBound();
        tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);

        zAcc = z.rightBound();
        tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }

    // intersection of `y = 0 && z = 0`
    if (isSingular(y) && isSingular(z))
    {
        xAcc = x.leftBound();
        yAcc = 0.0;
        zAcc = 0.0;
        tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);

        xAcc = x.rightBound();
        tempValues(xAcc, yAcc, zAcc, tempx, tempy, tempz, rExp, bExp);
        setMinMax(tempx, tempy, tempz, qx, qy, qz);
    }
}

