




void imageOfPointByReflectionByLine(MyInterval p, MyInterval q, MyInterval a, MyInterval b, MyInterval xk, MyInterval yk,
                                    MyInterval &xkbis, MyInterval &ykbis)
/** We have a line: (p,q) + t(a,b) with a^2 + b^2 = 1  (**)
 * We're looking for reflection of the point (xk, yk) WRT this line.
 * Resulting point is: (xkbis, ykbis)
 *
 */
{
    // vector perpendicular to the line (**): (-b, a)
    MyInterval c = scalarProduct(xk - p, yk - q, -b, a);
    xkbis = xk - 2.0 * c * (-b);
    ykbis = yk - 2.0 * c * a;
}

void computeBisector(MyInterval cosA, MyInterval sinA, MyInterval &cosB, MyInterval &sinB)
// we want to find a bisector of an angle given by a direction (cosA, sinA) and 0X axis i.e. (1,0)
// on output (cosB, sinB) is a direction of the bisector
{
    // we solve equations:
    // sinA = 2 * sinB *cosB
    // cosA = cosB^2 - sinB^2
	/////  for sinB and cosB

    MyInterval cosBSquare = (cosA + 1.0) / 2.0;
    MyInterval sinBSquare = (1.0 - cosA) / 2.0;

    if (cosBSquare > sinBSquare)
    {
        cosB = sqrt(cosBSquare);
        if (isSingular(cosB))
        {
            cout << "Problem with 'computeBisector', cosB" << endl;
            exit(1);
        }
        sinB = sinA / (2.0 * cosB);
    }
    else
    {
        sinB = sqrt(sinBSquare);
        if (isSingular(sinB))
        {
            cout << "Problem with 'computeBisector', sinB" << endl;
            exit(1);
        }
        cosB = sinA / (2.0 * sinB);
    }
}

bool possibleReflectionalSymmetry(const Bodies &bodies, int permutation[], int bodyA, MyInterval &sinB, MyInterval &cosB)
//  true is returned when we have found a posible line of the reflectional symmetry of <bodies>, which exchanges body <bodyA> with <n-2>-th
//     on output permutation[] how this works on points and (cosB,sinB) is the normalized direction of the reflection line passing through 0
{
    int N = bodies.N;

    // initiate permutation=Identity
    for (int i = 0; i < N; i++)
        permutation[i] = i;

   

    // finding parameters describing  bisector of the angle between q(N-2) and q(bodyA)
    MyInterval xA = bodies.x(bodyA);
    MyInterval yA = bodies.y(bodyA);
    MyInterval rA = sqrt(power(xA, 2) + power(yA, 2));
    MyInterval cosA = xA / rA;
    MyInterval sinA = yA / rA;

    // cosB, sinB -- cosinus and sinus of the direction of bisector line
    computeBisector(cosA, sinA, cosB, sinB);

    // Constructing permutation (beginning)
	permutation[bodyA] = N - 2;
    permutation[N - 2] = bodyA;
    MyInterval xkbis, ykbis;
    for (int k = 0; k <= N - 1; k++)
    {
        bool foundPair = false;

        if (k == bodyA || k == N - 2)
        {
            foundPair = true;
            continue;
        }

        MyInterval xk = bodies.x(k);
        MyInterval yk = bodies.y(k);

        imageOfPointByReflectionByLine(0, 0, cosB, sinB, xk, yk, xkbis, ykbis); // (0,0) because the line passes through the origin
        // does the body qk touch the symmetry line?
        if (!emptyIntersection(xk, xkbis) && !emptyIntersection(yk, ykbis)) // hence the permutation[k] = k
        {
            foundPair = true;
            continue;
        }

        for (int l = 0; l < N; l++)
        {
            if (l == k)
                continue; // qk and ql are the same body

            // check if the body ql can be a reflection of qk?
            MyInterval xl = bodies.x(l);
            MyInterval yl = bodies.y(l);

            if (!emptyIntersection(xl, xkbis) && !emptyIntersection(yl, ykbis))
            {
                permutation[l] = k;
                permutation[k] = l;
                foundPair = true;
                break;
            }
        }

        if (!foundPair)
            return false;
    }

    return true;
}

bool checkReflectionalSym(const Bodies &bodies, int bodyA)
// <bodyA> is an index of the body with the same distance to (0,0) as body q(N-2)
// we verify whether there is a reflectional symmetry mapping body q(bodyA) onto q(N-2)
// the symmetry line is a bisector of the angle between q(N-2) and q(bodyA)
//  uses global variables: zerosTestedFile and undecidedFile in a non-essential way - which should be 
//   removed
{
    int N = bodies.N;
    int permutation[N];
    MyInterval cosB, sinB; // cosinus and sinus of the direction of bisector line

    if (possibleReflectionalSymmetry(bodies, permutation, bodyA, sinB, cosB))
    // each body has a pair or lies on the symmetry line
    {
        cout << "reflectional symmetry permutation: ";
        for (int i = 0; i < N; i++)
            cout << permutation[i] << ", ";
        cout << endl;
        // permutation building (end)

        // we build a symmetrical set to check for (un)ambiguity (beginning)
        Bodies symBodies(bodies);

        MyInterval xi, xj;
        MyInterval yi, yj;
        MyInterval xjBis, yjBis;
        for (int i = 0; i < N - 1; i++)
        {
            // take the body qi and its reflection qj = q_permutation[j]
           
           
            xi = bodies.x(i);
            yi = bodies.y(i);
            xj = bodies.x(permutation[i]);
            yj = bodies.y(permutation[i]);
            // calculate the reflection by the line; it hits xj (pointed to by permutation)
            imageOfPointByReflectionByLine(0, 0, cosB, sinB, xj, yj, xjBis, yjBis);
			if(permutation[i]==bodyA)
			   yjBis=0;  // because the line was defined so that the body "bodyA" is mapped on to body "n-2", hence y(image)=0  

            symBodies.setX(i, intervalHull(xi, xjBis));
            symBodies.setY(i, intervalHull(yi, yjBis));  // the function setY for i=N-2 takes care about y_{n-2}=0
        }
        // build a symmetrical set to check for (un)ambiguity (end)

         
		// the trash variables used for compatybility with the call of searchCC
        ofstream outFileTest;
     	ofstream outFileFailed;  
        outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
        outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);
    	// end of trash
 
		 
		 
        Count zeroFound = searchCC(symBodies, false,outFileTest,outFileFailed); // testing stage
	                                  
		outFileTest.close();
        outFileFailed.close();							  

        if (isUnique(zeroFound)) return true;
        return blowUpAndProveUniqueSolution(symBodies);
    }
    return false;
}


bool findAllREflectionalSymmetries(const Bodies &bodies, const vector<int> &vec, int &symCount)
//  return true if CC in bodies has a reflectional symmetry with respect to line different from OX axis
//   on input: vec[0,\dots,] the group containing q(N-2) (with q(N-2) inclusive) of bodies which could have the same
//             distance from the origin as q(N-2) , we have v[0]=n-2
//    on output:  symCount - the number of lines of refectional symmetry   different from OX axis
{
    symCount = 0;

    // size of the group containing q(N-2) (with q(N-2) inclusive)
    const int s = vec.size();

    for (int i = 1; i < s; i++)  // 0-th element is N-2
    {
        // check symmetry by bisector of the angle between q(N-2) and q(coord)
        if (checkReflectionalSym(bodies,vec[i] ))
            symCount++;
    }

    if (symCount > 0)
    {
        cout << "number of symmetry lines other than 0X = " << symCount << endl;
        return true;
    }

    return false;
}


void imageByRot(MyInterval sinBeta, MyInterval cosBeta,
                MyInterval xk, MyInterval yk, MyInterval &xkbis, MyInterval &ykbis)
{
    xkbis = xk * cosBeta - yk * sinBeta;
    ykbis = xk * sinBeta + yk * cosBeta;
}

bool checkSymRotational(const Bodies &bodies, int bodyA, MyInterval &sinBeta, MyInterval &cosBeta, int permutation[])
//  we verify whether there exists a rotation such that body <bodyA> is  mapped onto (n-2)-th body 
//  if this is the case "true"  is returned
//  On output:
//    permutation - contains the permutation acting on bodies in this symmetry
//    cosBeta,sinBeta - cos and sin of the rotation angle (= angle of the bodyA to be mapped onto OX axis)
// we rotate our CC 'bodies' about the angle 'beta' and we obtain normalized CC -- say 'bodiesBis'
// we verify whether 'bodiesBis' is the same solution as 'bodies' (only with permutation od bodies)
{
    bool possibleRotationalSymmetry = true;

    int N = bodies.N;

    // finding an angle between (a,b) and 0X
    // a = r * cosBeta
    // b = r * sinBeta
    MyInterval a = bodies.x(bodyA);
    MyInterval b = bodies.y(bodyA);
    MyInterval r = sqrt(power(a, 2) + power(b, 2)); // r is a radius in polar coordinates of a body q = (a,b)
    cosBeta = a / r;
    sinBeta = b / r;    // beta is an angle in polar coordinates of a body q = (a,b)
    sinBeta = -sinBeta; // to move q = (a,b) to OX we have to rotate by -beta angle

    // Permutation building (beginning)
    for (int k = 0; k < N; k++)
    {
        bool foundPair = false;

        MyInterval xk = bodies.x(k);
        MyInterval yk = bodies.y(k);

        MyInterval xkbis, ykbis;
        imageByRot(-sinBeta, cosBeta, xk, yk, xkbis, ykbis);

        for (int l = 0; l < N; l++)
        {
            // check, whether body qk can transfer into ql,
            // i.e. could qk rotated by beta become ql?
            MyInterval xl = bodies.x(l);
            MyInterval yl = bodies.y(l);

            if (!emptyIntersection(xl, xkbis) && !emptyIntersection(yl, ykbis))
            {
                permutation[k] = l;
                foundPair = true;
                break;
            }
        }

        if (!foundPair)
        {
            possibleRotationalSymmetry = false;
            break;
        }
    }
    // Permutation building (end)

    if (possibleRotationalSymmetry)
    // each body has a pair 
    {
        // Building a rotated set (beginning)
        Bodies symBodies(bodies);

        MyInterval xi, xj;
        MyInterval yi, yj;
        MyInterval xjBis, yjBis;
        for (int i = 0; i <= N - 2; i++)
        {
            // take the body qi and its "twin" qj = qpermutation[j]
            // j can equal N-1, so we cannot assign qj!
            xi = bodies.x(i);
            yi = bodies.y(i);

            xj = bodies.x(permutation[i]);
            yj = bodies.y(permutation[i]);
            // calculate what q(i) should transfer to:
            imageByRot(sinBeta, cosBeta, xj, yj, xjBis, yjBis);

            symBodies.setX(i, intervalHull(xi, xjBis));
            if (i != N - 2)
                symBodies.setY(i, intervalHull(yi, yjBis));
            else // because the angle beta was chosen so that a q(...) is mapped onto OX axis
                symBodies.setY(i, 0.0);
        }
        // Building a rotated set (end)
         
		 
		 
 		 // the trash variables used for compatybility with the call of searchCC
        ofstream outFileTest;
	    ofstream outFileFailed;  
        outFileTest.open(zerosTestedFile, ios::binary | ios::out | ios::trunc);
        outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);
	    // end of trash

        
        Count zeroFound = searchCC(symBodies, false,outFileTest,outFileFailed); // testing stage
	                               

        outFileTest.close();
		outFileFailed.close();
        
        if (isUnique(zeroFound)) return true;
        return blowUpAndProveUniqueSolution(symBodies);
        
    }
    return false;
}


bool cardOneGroupsContainsZero(const Bodies &bodies, const vector<vector<int>> &vec)
/** Returns `false` if there exists a group `vec[i]` with cardinality 1 that
 * contains a body different than (0,0).
 */
{
    int sizeVec = vec.size();
    for (int i = 0; i < sizeVec; i++) // loops through each row of vy
    {
        int s = vec[i].size();

        if (s == 1)
        {
            int j = vec[i][0];
            if (!isSingular(bodies.x(j)) || !isSingular(bodies.y(j)))
                return false; // the group vy[i] with cardinality 1 contains body different than (0,0)
        }
    }
    return true;
}

int computeRotAngle(int s, MyInterval sinBeta, MyInterval cosBeta)
/**
 * beta is an angle of rotation, which maps point (a,b) to a line 0X
 * Returns a value > 0 if there exists k such that:
 *  * 2 <= k <= s
 *  * sinBeta = sin(Pi/k) and cosBeta = cos(Pi/k)
 * Returns the maximal such k.
 * Returns zero if no rotation by angle 2Pi/k (2 <= k <= s) transforms (x,y) into (a,b)
 * Returns -1 if there can be more than one k values with these properties.
 * (Horrible situation, which never should occur for the found CC-s).
 */
{
    int kmax = 0;
    int numGoodk = 0;

    // cosTemp and sinTemp are solutions of the system:
    //   $$ \left\{
    //   \begin{array}{lcl}
    //   a & = & x\cdot \cos \beta -y\cdot \sin \beta
    //   b & = & x\cdot \sin \beta + y\cdot \cos \beta. \end{array}
    //   \right.
    //   $$

    for (int k = 2; k <= s; k++)
    {
        MyInterval ang = (2 * piConst) / k;
        MyInterval sinTemp = sin(ang);
        MyInterval cosTemp = cos(ang);
        if ((!emptyIntersection(sinTemp, sinBeta)) &&
            (!emptyIntersection(cosTemp, cosBeta)))
        {
            if (k > kmax)
            {
                kmax = k;
                numGoodk++;
            }
        }
    }
    if (numGoodk > 1)
        kmax = -1;

    return kmax;
}

//==================================================
int findSymRotational(const Bodies &bodies, const vector<vector<int>> &vec)
/** On input
 * `vec[i]`: the vector containing bodies with equal distances from 0, i.e.:
 *     `vec[i][0] = i, vec[i][j] = k` => bodies `i` and `k` have equal distances from 0.
 * Returns kSym such that 2*Pi/kSym is the smallest possible angle of rotation.
 * kSym == 1 means that the only rotation is identity
 * kSym == -1 means that we do not have any conclusion.
 * This is most probably due to very "wide" data, which should never happen for a CC.
 */
{
    int N = bodies.N;
    int permutation[N];
    int properpermutation[N];
	
	//  we initiate properpermutaion=Id
    for (int i = 0; i < N; i++)
    {
        properpermutation[i] = i;
    }

    if (!cardOneGroupsContainsZero(bodies, vec))
    {
       // cout << "There exists a group with cardinality 1 that "
       //      << "contains body different than (0,0)" << endl;
        return false;
    }

    int k;
    int kSym = 1;

    MyInterval x = bodies.x(N - 2); // vec[N-2][0] = N-2
    MyInterval y = bodies.y(N - 2);
    const int s = vec[N - 2].size(); // the number of bodies with the same distance to (0,0) as q(N-2)

    for (int num = 1; num < s; num++) // 0-th body in the vec[N-2] group is q(N-2), i.e. bodies.q(vec[N-2][0]) = q(N-2)
    // in this loop we are looking for the point (a,b) which can be mapped to q(N-2)
    {
        MyInterval beta, sinBeta, cosBeta;                                        // we recalculate beta, to obtain an angle in the form 2*pi/k;
        if (checkSymRotational(bodies, vec[N-2][num], sinBeta, cosBeta, permutation)) // rotation by -beta brings (a,b) to axis 0X
        {
            k = computeRotAngle(s, sinBeta, cosBeta);
            if (k > kSym)
            {
                beta = 2 * piConst / k;
                kSym = k;
                for (int i = 0; i < N; i++)
                    properpermutation[i] = permutation[i];
            }
        }
    }
    if (kSym > 1)
    {
        cout << "\npermutation: ";
        for (int i = 0; i < N; i++)
            cout << properpermutation[i] << ", ";
        cout << endl;
    }
    return kSym;
}


void  writeDiffCCandTestSymmetry(int ccNo, vector<Bodies> &ccSolutions, vector<int> &markedSolutions)
//  ccNo - the number of entries in ccSolutions, markedSolutions
//      ccSolution[i] is treated as a representant of CC iff markedSolutions[i]==0
//   the function writes to the standard output: different CCs together with the information of their symmetries
//                  for !equalmasses it makes sense to test only the collinearity 
//                   otherwise we check all reflections with respect to a line and the rotation
//  the function uses global variables: mathFile,mathListFile
{

    int numBodies=ccSolutions[0].N;
    
    int symOXCounter = 0;  // the counter for CC with OX-reflectional symmetry
    int symColCounter = 0;  // the counter for CC which are collinear
    

    cout << endl;
    cout << "Different CC:" << endl;
    //mathFile << "Different CC:" << endl;

    int realCC = 0; // the counter for different CC
    int counter = 0; // the flag of symmetries by lines:
    // for each body is 1 if there is a symmetry or 0 if there is no

    for (int i = 0; i < ccNo; i++)
        if (markedSolutions[i] == 0)
        {
            realCC++;
            mathFile << "---------------------" << endl;
            mathFile << "position " << i << endl;
            ccSolutions[i].writeCCformatMath(mathFile,mathListFile);

            cout << "---------------------" << endl;
            cout << "position " << i << endl;
            ccSolutions[i].printData(true); // writes distances from 0
            cout << flush;

            writePotAndInertia(ccSolutions[i],equalMasses); // in the case of equal masses writes Moeckel potential

            if (colinearCC(ccSolutions[i]))
            {
                symColCounter++;
                cout << "collinear solution no " << symColCounter << endl;
            }
            if (equalMasses)
            {
                counter = 0;
                if (symmetricXCC(ccSolutions[i]))
                {
                    symOXCounter++;
                    counter = 1; // there is a symmetry with respect to OX
                    cout << "symmetric with respect to OX no " << symOXCounter << endl;
                }

                int symCount=0;

                // the vector containing bodies equidistant from 0
                // vec[i][0] = i, vec[i][j] = k <=> bodies i and k have the same distance from 0
                vector<vector<int>> vec;
                findGroups(ccSolutions[i], vec);

                findAllREflectionalSymmetries(ccSolutions[i], vec[numBodies-2], symCount);
                   

                cout << "\nnumber of reflectional symmetries = "
                     << symCount + counter << endl;

                int kmax = findSymRotational(ccSolutions[i], vec);
                if (kmax > 1)
                    cout << "A rotation by 2*Pi/" << kmax << " angle is a symmetry." << endl;
                else if (kmax == -1)
                    cout << "WE CANNOT MAKE CONCLUSION ABOUT THE ROTAIONAL SYMMETRY" << endl;
                else
                    cout << "There is NO a rotational symmetry" << endl;
            }
        }
    cout << endl;
    cout << "Number of different cc = " << realCC << endl;
}
